(* ------------------------------------------------------------------------ *)
(* ---------------- MDULO DE IMPLEMENTACION.  VIDEOCL -------------------- *)
(* ------------------------------------------------------------------------ *)

IMPLEMENTATION MODULE videocl;


FROM Storage IMPORT ALLOCATE,DEALLOCATE,Available;

FROM cadenas IMPORT Longitud,Comparar,Copiar,TCadena;

IMPORT IO;  (* Utilizada para visualizar. *)

IMPORT FIO;  (* Usada para el tratamiento de ficheros  *)

FROM Window IMPORT Clear;


TYPE
  TVideo=POINTER TO TVideoClub;         (* Tipo General. *)
  TSocios=POINTER TO TSocio;
  TPeliculas=POINTER TO TPelicula;

  TVideoClub=RECORD                     (* Tipo de la cabecera. *)
    socios:TSocios;
    peliculas:TPeliculas;
  END;

  TSocio=RECORD                         (* Nodo de un socio. *)
    numero:CARDINAL;
    nombre: TCadena;
    direccion: TCadena;
    telefono: TCadena;
    sig: TSocios;
  END;

  TAlquiler=POINTER TO TUsuario;

  TPelicula=RECORD                      (* Nodo de una pelicula. *)
    titulo:TCadena;
    copias:CARDINAL;
    sig:TPeliculas;
    alquiler:TAlquiler;
  END;

  TUsuario=RECORD                       (* Nodo de un alquiler. *)
    numero:CARDINAL;
    sig:TAlquiler;
  END;


(* Los algoritmos que se importan al programa estn comentados en el *)
(* mdulo de definicin.                                             *)


PROCEDURE CrearEstructura(VAR video:TVideo; VAR error:TError);
BEGIN
  IF video=NIL THEN
    IF Available(SIZE(TVideoClub)) THEN
      NEW(video);
      video^.socios:=NIL;
      video^.peliculas:=NIL;
    ELSE
      error:=NoHayEspacio;
    END;
  ELSE
    error:=YaExisteEstructura
  END;
END CrearEstructura;



(* Proc. Local. Devuelve TRUE si la estructura ha sido creada. *)
PROCEDURE EstructuraCreada(video:TVideo):BOOLEAN;
VAR
  dev:BOOLEAN;
BEGIN
  dev:=TRUE;
  IF video=NIL THEN
    dev:=FALSE;
  END;
  RETURN(dev);
END EstructuraCreada;


PROCEDURE NuevoSocio(VAR video:TVideo; VAR error:TError; nom:ARRAY OF CHAR;
                     tel:ARRAY OF CHAR; dir:ARRAY OF CHAR);

    (* Devuelve el ltimo nmero que ha sido asignado a un socio. *)
    (* Si no hay socios devuelve cero.                            *)
    PROCEDURE UltimoNumero(video:TVideo):CARDINAL;
    VAR
      aux:TSocios;      (* Recorre la lista de socios. *)
      dev:CARDINAL;     (* Devuelve el resultado de la funcin. *)
    BEGIN
      dev:=0;
      IF video^.socios<>NIL THEN (* Si hay socios. *)
        dev:=video^.socios^.numero;
      END;
      RETURN(dev);
    END UltimoNumero;


    (* Busca un socio en la estructura y devuelve un puntero a l. *)
    (* Si no lo encuentra devuelve el puntero a NIL.               *)
    PROCEDURE BuscaSocio(video:TVideo; nom:ARRAY OF CHAR):TSocios;
    VAR
      aux:TSocios;              (* Recorre la lista de socios. *)
      encontrado:BOOLEAN;       (* Para la busqueda cuando lo encuentra. *)
    BEGIN
      aux:=video^.socios;
      encontrado:=FALSE;
      WHILE (aux<>NIL) AND (NOT encontrado) DO
        IF Comparar(aux^.nombre,nom) THEN
          encontrado:=TRUE;
        ELSE
          aux:=aux^.sig;
        END;
      END;
      RETURN(aux);
    END BuscaSocio;


VAR
  aux,ptr:TSocios;
  UltNum:CARDINAL;  (* Almacena el ltimo nmero asignado a un socio. *)
BEGIN          (*NuevoSocio *)
  IF EstructuraCreada(video) THEN
    IF BuscaSocio(video,nom)=NIL THEN
      IF Available(SIZE(TSocio)) THEN
        NEW(aux);
        Copiar(aux^.nombre,nom);         (* En el examen se permite la asignacion
                                            24/05/00   *)
        Copiar(aux^.telefono,tel);
        Copiar(aux^.direccion,dir);
        UltNum:=UltimoNumero(video);
        aux^.numero:=UltNum+1;
        aux^.sig:=video^.socios;    (*Mantiene el orden decreciente en la lista
                                      de socios. El socio con mayor num. es el 1*)
        video^.socios:=aux;
      ELSE
        error:=NoHayEspacio;
      END;
    ELSE
      error:=SocioYaExiste;
    END;
  ELSE
    error:=EstructuraNoCreada;
  END;
END NuevoSocio;



(* Busca una pelcula. Si la encuentra devuelve un puntero a ella. *)
(* Si no la encuentra devuelve NIL. *)
PROCEDURE BuscaPelicula(video:TVideo;tit:ARRAY OF CHAR):TPeliculas;
VAR
  aux:TPeliculas;       (* Recorre la lista de pelculas. *)
  encontrado:BOOLEAN;   (* Para la bsqueda cuando la encuentra. *)
BEGIN
  aux:=video^.peliculas;
  encontrado:=FALSE;
  WHILE (aux<>NIL) AND (NOT encontrado) DO
    IF Comparar(aux^.titulo,tit) THEN
      encontrado:=TRUE;
    ELSE
      aux:=aux^.sig;
    END;
  END;
  RETURN(aux);
END BuscaPelicula;




PROCEDURE NuevaPelicula(VAR video:TVideo; VAR error:TError; tit:ARRAY OF CHAR;
                        cop:CARDINAL);
VAR
  aux:TPeliculas;       (* Nuevo nodo a insertar. *)
BEGIN
  IF EstructuraCreada(video) THEN
    aux:=BuscaPelicula(video,tit);
    IF aux<>NIL THEN                     (* Si la pelcula existe. *)
      aux^.copias:=aux^.copias+cop;
    ELSE                                 (* La pelcula no existe. *)
      IF Available(SIZE(TPelicula)) THEN (* Puede ser que no est o vacia. *)
        NEW(aux);
        Copiar(aux^.titulo,tit);
        aux^.copias:=cop;
        aux^.alquiler:=NIL;
        aux^.sig:=video^.peliculas;
        video^.peliculas:=aux;
      ELSE
        error:=NoHayEspacio;
      END;
    END;
  ELSE
    error:=EstructuraNoCreada;
  END;
END NuevaPelicula;



(* Busca un usuario por su nmero y devuelve un puntero a l. *)
(* Si no lo encuentra devuelve NIL.                           *)
PROCEDURE ExisteNumero(video:TVideo;numero:CARDINAL):TSocios;
VAR
  aux:TSocios;                  (* Recorre la lista de socios. *)
  encontrado:BOOLEAN;           (* Para la bsqueda cuando lo encuentra. *)
BEGIN
  aux:=video^.socios;
  encontrado:=FALSE;
  WHILE (aux<>NIL) AND (NOT encontrado) DO
    IF aux^.numero=numero THEN
      encontrado:=TRUE;
    ELSE                           (* Se puede hacer mas eficiente!...:-) *)
      aux:=aux^.sig;
    END;
  END;
  RETURN(aux);
END ExisteNumero;



PROCEDURE Alquila(VAR video:TVideo; VAR error:TError; num:CARDINAL;
                  tit:ARRAY OF CHAR);
VAR
  soc:TSocios;          (* Apuntar al socio si existe. *)
  pel:TPeliculas;       (* Apunta a la pelcula a alquilar si existe. *)
  aux:TAlquiler;        (* Nuevo nodo de alquiler. *)
BEGIN
  IF EstructuraCreada(video) THEN
    IF ExisteNumero(video,num)<>NIL THEN
      pel:=BuscaPelicula(video,tit);
      IF pel<>NIL THEN
        IF pel^.copias>0 THEN
          IF Available(SIZE(TUsuario)) THEN
            NEW(aux);
            aux^.numero:=num;
            aux^.sig:=pel^.alquiler;
            pel^.alquiler:=aux;
            pel^.copias:=pel^.copias-1;
          ELSE
            error:=NoHayEspacio;
          END;
        ELSE
          error:=PeliculaAgotada;
        END;
      ELSE
        error:=PeliculaNoExiste;
      END;
    ELSE
      error:=SocioNoExiste;
    END;
  ELSE
    error:=EstructuraNoCreada;
  END;
END Alquila;



PROCEDURE Devuelve(VAR video:TVideo; VAR error:TError; num:CARDINAL;
                   tit:ARRAY OF CHAR; dias:CARDINAL; VAR pagar:CARDINAL);

    (* Busca un socio en la lista de alquiler. *)
    (* Si lo encuentra devuelve alq al nodo de alquiler y ant al anterior. *)
    (* Si no lo encuentra devuelve alq a NIL. *)
    PROCEDURE BuscaAlquiler(lista:TAlquiler; num:CARDINAL; VAR alq,ant:TAlquiler);
    BEGIN
      alq:=lista;
      ant:=NIL;
      WHILE (alq<>NIL) AND (alq^.numero<>num) DO
        ant:=alq;
        alq:=alq^.sig;
      END;
    END BuscaAlquiler;

VAR
  pel:TPeliculas;       (* Recorre la lista de peliculas. *)
  alq,ant:TAlquiler;    (* Recorren la lista de alquiler uno tras otro. *)
  lista:TAlquiler;      (* Lista de alquiler. *)
BEGIN
  IF EstructuraCreada(video) THEN
    pel:=BuscaPelicula(video,tit);
    IF pel<>NIL THEN
      lista:=pel^.alquiler;
      BuscaAlquiler(lista,num,alq,ant);
      IF alq<>NIL THEN
        IF ant=NIL THEN  (* Borrar la primera. *)
          pel^.alquiler:=alq^.sig;
          DISPOSE(alq);
        ELSE             (* Borrar otra que no es la primera. *)
          ant^.sig:=alq^.sig;
          DISPOSE(alq);
        END;
        pel^.copias:=pel^.copias+1;
        pagar:=dias*300;
      ELSE
        error:=AlquilerNoExiste;
      END;
    ELSE
      error:=PeliculaNoExiste;
    END;
  ELSE
    error:=EstructuraNoCreada
  END;
END Devuelve;



PROCEDURE Visualizar(video:TVideo; VAR error:TError);
VAR
  bas:CHAR;       (* Pausas. *)
  cont:CARDINAL;  (* Contador para hacer pausas en la visualizacin. *)
  soc:TSocios;    (* Recorre la lista de socios. *)
  pel:TPeliculas; (* Recorre la lista de pelculas. *)
  alq:TAlquiler;  (* Recorre la lista de alquiler. *)
BEGIN
  IF EstructuraCreada(video) THEN
    Clear;
    IO.WrLn;
    IO.WrStr("--- SOCIOS ---");
    IF video^.socios<>NIL THEN                   (* Imprimo socios. *)
      soc:=video^.socios;
      WHILE soc<>NIL DO
          IO.WrLn;IO.WrLn;
          IO.WrStr("Socio nmero: ");
          IO.WrCard(soc^.numero,0);
          IO.WrLn;
          IO.WrStr("Nombre: ");
          IO.WrStr(soc^.nombre);
          IO.WrLn;
          IO.WrStr("Telfono: ");
          IO.WrStr(soc^.telefono);
          IO.WrLn;
          IO.WrStr("Direccin: ");
          IO.WrStr(soc^.direccion);
          soc:=soc^.sig;
          IO.WrLn;
          IO.WrStr("Pulse tecla para  continuar la visualizacin. ");
          IO.RdLn;IO.RdLn;
          bas:=IO.RdKey();IO.RdLn;
      END;
    ELSE
      IO.WrLn;
      IO.WrStr("No hay socios registrados. ");
    END;
    IO.WrLn; IO.WrLn;                   (* Visualiza las pelculas. *)
    IO.WrStr("--- PELCULAS ---");
    IF video^.peliculas<>NIL THEN
      pel:=video^.peliculas;
      WHILE pel<>NIL DO
          IO.WrLn; IO.WrLn;
          IO.WrStr("Ttulo: ");
          IO.WrStr(pel^.titulo);
          IO.WrLn;
          IO.WrStr("Copias disponibles: ");
          IO.WrCard(pel^.copias,0);
          IO.WrLn;
          IO.WrStr("Lista de alquiler de esta pelcula: ");
          IO.WrLn;
          alq:=pel^.alquiler;         (* Visualiza la lista de alquiler. *)
          IF alq=NIL THEN
            IO.WrStr("Ningn socio tiene esta pelcula. ")
          END;
          cont:=1;
          WHILE alq<>NIL DO
            IF cont<4 THEN
              IO.WrStr("Socio: ");
              IO.WrCard(alq^.numero,0);
              IO.WrLn;
              alq:=alq^.sig;
            ELSE
              IO.WrLn;
              IO.WrStr("Pulse tecla para  continuar la visualizacin. ");
              IO.RdLn;IO.RdLn;
              bas:=IO.RdKey();IO.RdLn;
              cont:=1;
            END;
          END;
          pel:=pel^.sig;
          IO.WrLn;
          IO.WrStr("Pulse tecla para  continuar la visualizacin. ");
          IO.RdLn;IO.RdLn;
          bas:=IO.RdKey();IO.RdLn;
        END;
    ELSE
      IO.WrLn;
      IO.WrStr("No existen pelculas. ");
    END;
  ELSE
    error:=EstructuraNoCreada;
  END;
END Visualizar;


(* PROC. Local *)
(* Guarda los socios en al comienzo del fichero "video.dat" *)
PROCEDURE GuardarSocios(video:TVideo);
VAR
  mnj:FIO.File;      (* MaNeJador del fichero. *)
  soc:TSocios;       (* Recorre la lista de socios. *)
  num:CARDINAL;

  PROCEDURE NumeroDeSocios(soc:TSocios):CARDINAL;
  VAR cont:CARDINAL;
  BEGIN
    cont:=0;
    WHILE soc<>NIL DO
       soc:=soc^.sig;
       INC(cont);
    END;
    RETURN cont;
  END NumeroDeSocios;

BEGIN
    mnj:=FIO.Create("video.dat");              (* OJO a la apertura !!! *)
    soc:=video^.socios;
    num:=NumeroDeSocios(soc);
    FIO.WrCard(mnj,num,0);
    FIO.WrLn(mnj);
    FIO.WrStr(mnj,"SOCIOS");
    FIO.WrLn(mnj);
    WHILE soc<>NIL DO
      FIO.WrCard(mnj,soc^.numero,0);
      FIO.WrChar(mnj," ");
      FIO.WrStr(mnj,soc^.nombre);
      FIO.WrLn(mnj);
      FIO.WrStr(mnj,soc^.telefono);
      FIO.WrLn(mnj);
      FIO.WrStr(mnj,soc^.direccion);
      FIO.WrLn(mnj);
      soc:=soc^.sig;
    END;
    FIO.Close(mnj);
END GuardarSocios;

(* PROC Local *)
(* Guarda las peliculas en el fichero "Video.dat" *)
PROCEDURE GuardarPeliculas(video:TVideo);
VAR
  num:CARDINAL;
  mnj:FIO.File;         (* MaNeJador del fichero. *)
  pel:TPeliculas;       (* Recorre la lista de peliculas. *)
  alq:TAlquiler;        (* Recorre la lista de alquileres. *)

  PROCEDURE NumeroDePeliculas(alq:TAlquiler):CARDINAL;
  VAR cont:CARDINAL;
  BEGIN
    cont:=0;
    WHILE alq<>NIL DO
       alq:=alq^.sig;
       INC(cont);
    END;
    RETURN cont;
  END NumeroDePeliculas;

BEGIN
    mnj:=FIO.Append("video.dat");       (* Ojo a la apertura *)
    pel:=video^.peliculas;
    FIO.WrStr(mnj,"PELICULAS");
    FIO.WrLn(mnj);
    WHILE pel<>NIL DO
      FIO.WrStr(mnj,pel^.titulo);
      FIO.WrLn(mnj);
      FIO.WrCard(mnj,pel^.copias,0);
      FIO.WrChar(mnj," ");
      FIO.WrStr(mnj,"En alquiler: ");
      FIO.WrLn(mnj);
      alq:=pel^.alquiler;
      num:=NumeroDePeliculas(alq);
      FIO.WrCard(mnj,num,0);
      FIO.WrChar(mnj," ");
      FIO.WrStr(mnj,"Copias");
      FIO.WrLn(mnj);
      WHILE alq<>NIL DO
        FIO.WrCard(mnj,alq^.numero,0);
        FIO.WrLn(mnj);
        alq:=alq^.sig;
      END;
      pel:=pel^.sig;
    END;
    FIO.Close(mnj);
END GuardarPeliculas;


PROCEDURE VolcarTexto(video:TVideo; VAR error:TError);
BEGIN
  IF EstructuraCreada(video) THEN
      GuardarSocios(video);
      GuardarPeliculas(video);
  ELSE
    error:=EstructuraNoCreada;
  END;
END VolcarTexto;



PROCEDURE Regalo(VAR video:TVideo; VAR error:TError; tit:ARRAY OF CHAR);
VAR
  pel:TPeliculas;   (* Recorre la lista de pelculas. *)
  ant:TPeliculas;   (* Anterior a pel. *)
  alq:TAlquiler;    (* Recorre la lista de alquileres. *)
BEGIN
  IF EstructuraCreada(video) THEN
    pel:=BuscaPelicula(video,tit);
    IF pel<>NIL THEN
      IF pel^.alquiler<>NIL THEN
        WHILE pel^.alquiler<>NIL DO  (* Borro los alquileres sin reponer. *)
          alq:=pel^.alquiler;
          pel^.alquiler:=pel^.alquiler^.sig;
          DISPOSE(alq);
        END;
        IF pel^.copias=0 THEN  (* Hay que borrar la pelcula. *)
          IF video^.peliculas=pel THEN  (* Borrar la primera. *)
            video^.peliculas:=video^.peliculas^.sig;
            DISPOSE(pel);
          ELSE                          (* Borrar una que no es primera. *)
            ant:=video^.peliculas;      (* Busco la anterior. *)
            WHILE ant^.sig<>pel DO
              ant:=ant^.sig;
            END;
            ant^.sig:=pel^.sig;
            DISPOSE(pel);
          END;
        END;
      ELSE
        error:=NadieAlquila;
      END;
    ELSE
      error:=PeliculaNoExiste;
    END;
  ELSE
    error:=EstructuraNoCreada;
  END;
END Regalo;

PROCEDURE FichATAD(VAR video:TVideo;VAR error:TError);
VAR
  mnj:FIO.File;      (* MaNeJador del fichero. *)
  soc,aux:TSocios;       (* Recorre la lista de socios. *)
  numero,num,i:CARDINAL;
  ch:CHAR;
  cad,nombre,direccion,telefono:TCadena;
  pel:TPeliculas;       (* Recorre la lista de peliculas. *)
  alq:TAlquiler;        (* Recorre la lista de alquileres. *)

BEGIN
    Destruir(video,error);
    CrearEstructura(video,error);
    mnj:=FIO.Open("video.dat");              (* OJO a la apertura !!! *)

    num:=FIO.RdCard(mnj);
    FIO.RdStr(mnj,cad);
    FOR i:=1 TO num DO
      NEW(soc);
      soc^.numero:=FIO.RdCard(mnj);
      FIO.RdStr(mnj,soc^.nombre);
      FIO.RdStr(mnj,soc^.telefono);
      FIO.RdStr(mnj,soc^.direccion);
      IF video^.socios=NIL THEN    (*Ahora insertamos de mayor a menor *)
         video^.socios:=soc
      ELSE
         aux:=video^.socios;
         WHILE aux^.sig<>NIL DO
           aux:=aux^.sig;
         END;
         aux^.sig:=soc;
      END;
    END;
    FIO.RdStr(mnj,cad);
    FIO.RdStr(mnj,cad);
    WHILE NOT FIO.EOF DO
      NEW(pel);
      pel^.titulo:=cad;
      pel^.copias:=FIO.RdCard(mnj);
      FIO.RdStr(mnj,cad);
      pel^.sig:=video^.peliculas;
      video^.peliculas:=pel;

      num:=FIO.RdCard(mnj);
      FIO.RdStr(mnj,cad);
      pel^.alquiler:=NIL;
      FOR i:=1 TO num DO
          NEW(alq);
          alq^.numero:=FIO.RdCard(mnj);
          alq^.sig:=pel^.alquiler;
          pel^.alquiler:=alq;
      END;
      FIO.RdStr(mnj,cad);
    END;
    FIO.Close(mnj);
END FichATAD;



PROCEDURE Destruir(VAR video:TVideo; VAR error:TError);
VAR
  soc:TSocios;     (* Recorre la lista de socios. *)
  pel:TPeliculas;  (* Recorre la lista de pelculas. *)
  alq:TAlquiler;   (* Recorre la lista de alquiler. *)
BEGIN;
  IF EstructuraCreada(video) THEN
    WHILE (video^.peliculas<>NIL) DO                 (* Borro pelculas. *)
      WHILE (video^.peliculas^.alquiler<>NIL) DO     (* Borro alquiler. *)
        alq:=video^.peliculas^.alquiler;
        video^.peliculas^.alquiler:=video^.peliculas^.alquiler^.sig;
        DISPOSE(alq);
      END;
      pel:=video^.peliculas;
      video^.peliculas:=video^.peliculas^.sig;
      DISPOSE(pel);
    END;
    WHILE (video^.socios<>NIL) DO                   (* Borro socios. *)
      soc:=video^.socios;
      video^.socios:=video^.socios^.sig;
      DISPOSE(soc);
    END;
    DISPOSE(video);                                 (* Borro cabecera. *)
  ELSE
    error:=NoHay;
  END;
END Destruir;



END videocl.
